from os.path import abspath, realpath

from nose.tools import nottest
import nose.util


@nottest
def test_address(test):
    """Return the result of nose's test_address(), None if it's stumped."""
    try:
        return nose.util.test_address(test)
    except TypeError:  # Explodes if the function passed to @with_setup applied
                       # to a test generator has an error.
        pass


def nose_selector(test):
    """Return the string you can pass to nose to run `test`.

    Return "Unknown test" if it can't construct a decent path.

    """
    address = test_address(test)
    if address:
        file, module, rest = address

        if module:
            if rest:
                return '%s:%s' % (module, rest)
            else:
                return module
    return 'Unknown test'


class OneTrackMind(object):
    """An accurate simulation of my brain

    I can know one thing at a time, at some level of confidence. You can tell
    me other things, but if I'm not as confident of them, I'll forget them. If
    I'm more confident of them, they'll replace what I knew before.

    """
    def __init__(self):
        self.confidence = 0
        self.best = None

    def know(self, what, confidence):
        """Know something with the given confidence, and return self for chaining.

        If confidence is higher than that of what we already know, replace
        what we already know with what you're telling us.

        """
        if confidence > self.confidence:
            self.best = what
            self.confidence = confidence
        return self


@nottest  # still needed?
def index_of_test_frame(extracted_tb, exception_type, exception_value, test):
    """Return the index of the frame that points to the failed test or None.

    Sometimes this is hard. It takes its best guess. If exception_type is
    SyntaxError or it has no idea, it returns None.

    Args:
        address: The result of a call to test_address(), indicating which test
            failed
        exception_type, exception_value: Needed in case this is a SyntaxError
            and therefore doesn't have the whole story in extracted_tb
        extracted_tb: The traceback, after having been passed through
            extract_tb()

    """
    try:
        address = test_address(test)
    except TypeError:
        # Explodes if the function passed to @with_setup
        # applied to a test generator has an error.
        address = None

    # address is None if the test callable couldn't be found. No sense trying
    # to find the test frame if there's no such thing:
    if address is None:
        return None

    test_file, _, test_call = address

    # OneTrackMind helps us favor the latest frame, even if there's more than
    # one match of equal confidence.
    knower = OneTrackMind()

    if isinstance(test_file, basestring):  # Sometimes it's None.
        test_file_path = realpath(test_file)

        # TODO: Perfect. Right now, I'm just comparing by function name within
        # a module. This should break only if you have two identically-named
        # functions from a single module in the call stack when your test
        # fails. However, it bothers me. I'd rather be finding the actual
        # callables and comparing them directly, but that might not work with
        # test generators.
        for i, frame in enumerate(extracted_tb):
            file, line, function, text = frame
            if file is not None and test_file_path == realpath(file):
                # TODO: Now that we're eliding until the test frame, is it
                # desirable to have this confidence-2 guess when just the file
                # path is matched?
                knower.know(i, 2)
                if (hasattr(test_call, 'rsplit') and  # test_call can be None
                    function == test_call.rsplit('.')[-1]):
                    knower.know(i, 3)
                    break
    return knower.best


def human_path(path, cwd):
    """Return the most human-readable representation of the given path.

    If an absolute path is given that's within the current directory, convert
    it to a relative path to shorten it. Otherwise, return the absolute path.

    """
    # TODO: Canonicalize the path to remove /kitsune/../kitsune nonsense.
    path = abspath(path)
    if cwd and path.startswith(cwd):
        path = path[len(cwd) + 1:]  # Make path relative. Remove leading slash.
    return path



"""Fancy traceback formatting"""

import os
from traceback import extract_tb, format_exception_only

from blessings import Terminal
from nose.util import src

from noseprogressive.utils import human_path


def format_traceback(extracted_tb,
                     exc_type,
                     exc_value,
                     cwd='',
                     term=None,
                     function_color=12,
                     dim_color=8,
                     editor='vi'):
    """Return an iterable of formatted traceback frames.

    Also include a pseudo-frame at the end representing the exception itself.

    Format things more compactly than the stock formatter, and make every
    frame an editor shortcut.

    """
    def format_shortcut(editor,
                        file,
                        line,
                        function=None):
        """Return a pretty-printed editor shortcut."""
        return template % dict(editor=editor,
                               line=line,
                               file=file,
                               function=('  # ' + function) if function else '',
                               funcemph=term.color(function_color),
                               # Underline is also nice and doesn't make us
                               # worry about appearance on different background
                               # colors.
                               plain=term.normal,
                               fade=term.color(dim_color) + term.bold)

    if not term:
        term = Terminal()

    # Shorten file paths:
    for i, (file, line, function, text) in enumerate(extracted_tb):
        extracted_tb[i] = human_path(src(file), cwd), line, function, text

    line_width = len(str(max(the_line for _, the_line, _, _ in extracted_tb)))
    template = ('  %(fade)s%(editor)s +%(line)-' + str(line_width) + 's '
                '%(file)s%(plain)s'
                '%(funcemph)s%(function)s%(plain)s\n')

    # Stack frames:
    for i, (file, line, function, text) in enumerate(extracted_tb):
        yield (format_shortcut(editor, file, line, function) +
               ('    %s\n' % (text or '')))

    # Exception:
    if exc_type is SyntaxError:
        # Format a SyntaxError to look like our other traceback lines.
        # SyntaxErrors have a format different from other errors and include a
        # file path which looks out of place in our newly highlit, editor-
        # shortcutted world.
        exc_lines = [format_shortcut(editor, exc_value.filename, exc_value.lineno)] + \
                     format_exception_only(SyntaxError, exc_value)[1:]
    else:
        exc_lines = format_exception_only(exc_type, exc_value)
    yield ''.join(exc_lines)


# Adapted from unittest:

def extract_relevant_tb(tb, exctype, is_test_failure):
    """Return extracted traceback frames that aren't unittest ones.

    This used to be _exc_info_to_string().

    """
    # Skip test runner traceback levels:
    while tb and _is_unittest_frame(tb):
        tb = tb.tb_next
    if is_test_failure:
        # Skip assert*() traceback levels:
        length = _count_relevant_tb_levels(tb)
        return extract_tb(tb, length)
    return extract_tb(tb)


def _is_unittest_frame(tb):
    """Return whether the given frame is something other than a unittest one."""
    return '__unittest' in tb.tb_frame.f_globals


def _count_relevant_tb_levels(tb):
    """Return the number of frames in ``tb`` before all that's left is unittest frames.

    Unlike its namesake in unittest, this doesn't bail out as soon as it hits a
    unittest frame, which means we don't bail out as soon as somebody uses the
    mock library, which defines ``__unittest``.

    """
    length = contiguous_unittest_frames = 0
    while tb:
        length += 1
        if _is_unittest_frame(tb):
            contiguous_unittest_frames += 1
        else:
            contiguous_unittest_frames = 0
        tb = tb.tb_next
    return length - contiguous_unittest_frames



"""Tests for traceback formatting.

At the moment, we test only that it doesn't crash.

"""
from noseprogressive.tracebacks import format_traceback


syntax_error_tb = ([
     ("setup.py", 79, '?', """classifiers = ["""),
     ("/usr/lib64/python2.4/distutils/core.py", 149, 'setup', """dist.run_commands()"""),
     ("/usr/lib64/python2.4/distutils/dist.py", 946, 'run_commands', """self.run_command(cmd)"""),
     ("/usr/lib64/python2.4/distutils/dist.py", 966, 'run_command', """cmd_obj.run()"""),
     ("/usr/lib/python2.4/site-packages/setuptools/command/install.py", 76, 'run', """self.do_egg_install()"""),
     ("/usr/lib/python2.4/site-packages/setuptools/command/install.py", 100, 'do_egg_install', """cmd.run()"""),
     ("/usr/lib/python2.4/site-packages/setuptools/command/easy_install.py", 211, 'run', """self.easy_install(spec, not self.no_deps)"""),
     ("/usr/lib/python2.4/site-packages/setuptools/command/easy_install.py", 427, 'easy_install', """return self.install_item(None, spec, tmpdir, deps, True)"""),
     ("/usr/lib/python2.4/site-packages/setuptools/command/easy_install.py", 473, 'install_item', """self.process_distribution(spec, dist, deps)"""),
     ("/usr/lib/python2.4/site-packages/setuptools/command/easy_install.py", 518, 'process_distribution', """distros = WorkingSet([]).resolve("""),
     ("/usr/lib/python2.4/site-packages/pkg_resources.py", 481, 'resolve', """dist = best[req.key] = env.best_match(req, self, installer)"""),
     ("/usr/lib/python2.4/site-packages/pkg_resources.py", 717, 'best_match', """return self.obtain(req, installer) # try and download/install"""),
     ("/usr/lib/python2.4/site-packages/pkg_resources.py", 729, 'obtain', """return installer(requirement)"""),
     ("/usr/lib/python2.4/site-packages/setuptools/command/easy_install.py", 432, 'easy_install', """dist = self.package_index.fetch_distribution("""),
     ("/usr/lib/python2.4/site-packages/setuptools/package_index.py", 462, 'fetch_distribution', """self.find_packages(requirement)"""),
     ("/usr/lib/python2.4/site-packages/setuptools/package_index.py", 303, 'find_packages', """self.scan_url(self.index_url + requirement.unsafe_name+'/')"""),
     ("/usr/lib/python2.4/site-packages/setuptools/package_index.py", 612, 'scan_url', """self.process_url(url, True)"""),
     ("/usr/lib/python2.4/site-packages/setuptools/package_index.py", 190, 'process_url', """f = self.open_url(url)"""),
     ("/usr/lib/python2.4/site-packages/setuptools/package_index.py", 579, 'open_url', """return open_with_auth(url)"""),
     ("/usr/lib/python2.4/site-packages/setuptools/package_index.py", 676, 'open_with_auth', """fp = urllib2.urlopen(request)"""),
     ("/usr/lib64/python2.4/urllib2.py", 130, 'urlopen', """return _opener.open(url, data)"""),
     ("/usr/lib64/python2.4/urllib2.py", 358, 'open', """response = self._open(req, data)"""),
     ("/usr/lib64/python2.4/urllib2.py", 376, '_open', """'_open', req)"""),
     ("/usr/lib64/python2.4/urllib2.py", 337, '_call_chain', """result = func(*args)"""),
     ("/usr/lib64/python2.4/urllib2.py", 573, '<lambda>', """lambda r, proxy=url, type=type, meth=self.proxy_open: \\"""),
     ("/usr/lib64/python2.4/urllib2.py", 580, 'proxy_open', """if '@' in host:""")
     # Was originally TypeError: iterable argument required
    ], SyntaxError, SyntaxError('invalid syntax', ('/Users/erose/Checkouts/nose-progress/noseprogressive/tests/test_integration.py', 97, 5, '    :bad\n')))
attr_error_tb = ([
     ("/usr/share/PackageKit/helpers/yum/yumBackend.py", 2926, 'install_signature', """self.yumbase.getKeyForPackage(pkg, askcb = lambda x, y, z: True)"""),
     ("/usr/lib/python2.6/site-packages/yum/__init__.py", 4309, 'getKeyForPackage', """result = ts.pgpImportPubkey(misc.procgpgkey(info['raw_key']))"""),
     ("/usr/lib/python2.6/site-packages/rpmUtils/transaction.py", 59, '__getattr__', """return self.getMethod(attr)"""),
     ("/usr/lib/python2.6/site-packages/rpmUtils/transaction.py", 69, 'getMethod', """return getattr(self.ts, method)""")
    ], AttributeError, AttributeError("'NoneType' object has no attribute 'pgpImportPubkey'"))


def test_syntax_error():
    """Exercise special handling of syntax errors."""
    ''.join(format_traceback(*syntax_error_tb))


def test_non_syntax_error():
    """Exercise typical error formatting."""
    ''.join(format_traceback(*attr_error_tb))
